{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 15](https://adventofcode.com/2019/day/15) puzzle input is an IntCode program (mine is [here](puzzle_input/day15.txt)) which controls a droid that should perform some repairs on an oxygen tank. Part 1 involves using the program to explore the surroundings (it's basically a maze), find a destination (an oxygen tank) then find the shortest path to that location. Part 2 involves exploring the maze, then simulating the spread of oxygen through the maze and finding how many time-steps it would take to fill every spot of the maze."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.done = False\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "           \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "        self.stalled = False\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "\n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "226\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<img src=\"./img/day15-part1.png\"/>"
      ],
      "text/plain": [
       "<IPython.core.display.Image object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "from PIL import Image, ImageDraw\n",
    "from IPython.display import Image as IPythonImage\n",
    "from IPython.display import display\n",
    "\n",
    "code = open(\"puzzle_input/day15.txt\", \"r\").read().strip().split(\",\")\n",
    "\n",
    "direction_north = 1\n",
    "direction_south = 2\n",
    "direction_west  = 3\n",
    "direction_east  = 4\n",
    "\n",
    "reverse_direction = {\n",
    "    direction_north: direction_south,\n",
    "    direction_south: direction_north,\n",
    "    direction_west: direction_east,\n",
    "    direction_east: direction_west\n",
    "}\n",
    "\n",
    "def backwards(movement_stack):\n",
    "    return [ reverse_direction[movement] for movement in movement_stack ]\n",
    "\n",
    "result_fail = 0\n",
    "result_ok = 1\n",
    "result_completed = 2\n",
    "\n",
    "def neighbours(position, movement_stack):\n",
    "    x, y = position\n",
    "    return [\n",
    "        ((x, y+1), movement_stack + [ direction_north ]),\n",
    "        ((x, y-1), movement_stack + [ direction_south ]),\n",
    "        ((x+1, y), movement_stack + [ direction_west ]),\n",
    "        ((x-1, y), movement_stack + [ direction_east ]),\n",
    "    ]\n",
    "    \n",
    "    \n",
    "\n",
    "def create_image(open_points, wall_points, filename):\n",
    "    if not filename:\n",
    "        return\n",
    "    \n",
    "    min_x = min([ x for x,y in open_points + wall_points ] + [0])\n",
    "    min_y = min([ y for x,y in open_points + wall_points ] + [0])\n",
    "    max_x = max([ x for x,y in open_points + wall_points ] + [100])\n",
    "    max_y = max([ y for x,y in open_points + wall_points ] + [100])\n",
    "    \n",
    "    offset_x = -(min_x)\n",
    "    offset_y = -(min_y)\n",
    "    \n",
    "    img = Image.new(\"RGB\", (max_x + offset_x, max_y + offset_y), \"black\")\n",
    "    draw = ImageDraw.Draw(img)\n",
    "    draw.point([ (x + offset_x, y + offset_y) for x,y in open_points], fill=\"white\")\n",
    "    draw.point([ (x + offset_x, y + offset_y) for x,y in wall_points], fill=\"red\")\n",
    "    img.save(filename)\n",
    "\n",
    "\n",
    "def check_contents_after_moves(code, movement_stack):\n",
    "    cpu = IntCodeCpu(code)\n",
    "    output_buffer = []\n",
    "    input_buffer = [ m for m in movement_stack ][::-1]\n",
    "    cpu.start(input_buffer, output_buffer)\n",
    "    return int(cpu.output_buffer[-1])    \n",
    "\n",
    "def solve_maze(filename):\n",
    "    queue = neighbours([0, 0], [])\n",
    "    visited = set([(0, 0)])\n",
    "    \n",
    "    # this is just to create the pretty picture at the end\n",
    "    open_points = [(0,0)]\n",
    "    wall_points = []\n",
    "    \n",
    "    while queue:\n",
    "        position, movement_stack = queue.pop(0)               \n",
    "        if position in visited:\n",
    "            continue\n",
    "            \n",
    "        visited.add(position)          \n",
    "        res = check_contents_after_moves(code, movement_stack)\n",
    "        \n",
    "        if res == result_completed:\n",
    "            create_image(open_points, wall_points, filename)\n",
    "            return movement_stack\n",
    "            \n",
    "        if res == result_fail:\n",
    "            wall_points.append(position)\n",
    "            \n",
    "        if res == result_ok:\n",
    "            open_points.append(position)\n",
    "            queue.extend([ (p,s) for p,s in neighbours(position, movement_stack) if not (p in visited)])\n",
    "    \n",
    "    raise Exception(\"No route to the stupid oxygen tank\")\n",
    "        \n",
    "filename = \"./img/day15-part1.png\"\n",
    "print(len(solve_maze(filename)))\n",
    "display(IPythonImage(url=filename))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "342\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<img src=\"./img/day15-part2.png\"/>"
      ],
      "text/plain": [
       "<IPython.core.display.Image object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "def check_contents_after_moves_2(code, movement_stack, initialisation):\n",
    "    cpu = IntCodeCpu(code)\n",
    "    output_buffer = []\n",
    "    input_buffer = [ m for m in initialisation + movement_stack ][::-1]\n",
    "    cpu.start(input_buffer, output_buffer)\n",
    "    return int(cpu.output_buffer[-1])\n",
    "\n",
    "\n",
    "def flood_maze(filename, input_instructions):   \n",
    "    queue = [ neighbours([0, 0], []) ]\n",
    "    visited = set([(0, 0)])\n",
    "    \n",
    "    # this is just to create the pretty picture at the end\n",
    "    open_points = [(0,0)]\n",
    "    wall_points = []\n",
    "    \n",
    "    minutes = 0\n",
    "    while queue:\n",
    "        minutes += 1\n",
    "        locations_group = queue.pop(0)\n",
    "        to_add_to_queue = []\n",
    "        \n",
    "        for position, movement_stack in locations_group:\n",
    "        \n",
    "            if position in visited:\n",
    "                continue\n",
    "\n",
    "            visited.add(position)          \n",
    "            res = check_contents_after_moves_2(code, movement_stack, input_instructions)\n",
    "\n",
    "            if res == result_fail:\n",
    "                wall_points.append(position)\n",
    "\n",
    "            if res == result_ok:\n",
    "                open_points.append(position)\n",
    "                to_add_to_queue.extend([ (p,s) for p,s in neighbours(position, movement_stack) if not (p in visited)])\n",
    "    \n",
    "        if to_add_to_queue:\n",
    "            queue.append(to_add_to_queue)\n",
    "            \n",
    "    create_image(open_points, wall_points, filename)\n",
    "    return minutes\n",
    "            \n",
    "filename = \"./img/day15-part2.png\"\n",
    "path_to_oxygen = solve_maze(None)\n",
    "print(flood_maze(filename, path_to_oxygen))\n",
    "display(IPythonImage(url=filename))"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
